@@ -6,6 +6,7 @@ import configparser
 
 
 from kinto import main as kinto_main
+from kinto.core import errors
 from kinto.core.events import AfterResourceChanged
 from kinto.core.testing import BaseWebTest, get_user_headers, FormattedErrorMixin
 from kinto_emailer import get_messages, send_notification
@@ -39,8 +40,9 @@ class EmailerTest(BaseWebTest, unittest.TestCase):
     api_prefix = "v1"
     config = 'config/kinto.ini'
 
-    def get_app_settings(self, extras=None):
-        ini_path = os.path.join(HERE, self.config)
+    @classmethod
+    def get_app_settings(cls, extras=None):
+        ini_path = os.path.join(HERE, cls.config)
         config = configparser.ConfigParser()
         config.read(ini_path)
         settings = dict(config.items('app:main'))
@@ -155,6 +157,28 @@ class GetMessagesTest(unittest.TestCase):
 
         assert len(messages) == 0
 
+    def test_get_messages_can_filter_using_regexps(self):
+        self.storage.get.return_value = {
+            'kinto-emailer': {
+                'hooks': [{
+                    'collection_id': '^(?!normandy-recipes$)',
+                    'template': 'Poll changed.',
+                    'recipients': ['me@you.com'],
+                }]
+            }
+        }
+        self.payload.update({'collection_id': 'normandy-recipes'})
+        messages = get_messages(self.storage, self.payload)
+        assert len(messages) == 0
+
+        self.payload.update({'collection_id': 'some-normandy-recipes'})
+        messages = get_messages(self.storage, self.payload)
+        assert len(messages) == 1
+
+        self.payload.update({'collection_id': 'normandy-recipes-all'})
+        messages = get_messages(self.storage, self.payload)
+        assert len(messages) == 1
+
     def test_get_messages_ignores_resource_if_not_specified(self):
         self.storage.get.return_value = {
             'kinto-emailer': {
@@ -209,11 +233,30 @@ class GetMessagesTest(unittest.TestCase):
         assert len(messages) == 1
 
 
+class BucketTest(unittest.TestCase):
+
+    def test_hooks_can_be_defined_on_buckets(self):
+        storage = mock.MagicMock()
+        collection_metadata = {}
+        bucket_metadata = {
+            'kinto-emailer': {
+                'hooks': [{
+                    'template': 'Poll changed.',
+                    'recipients': ['from@bucket.com'],
+                }]
+            }
+        }
+        storage.get.side_effect = [collection_metadata, bucket_metadata]
+        payload = {'bucket_id': 'b', 'collection_id': 'c'}
+        message, = get_messages(storage, payload)
+        assert message.recipients == ['from@bucket.com']
+
+
 class GroupExpansionTest(unittest.TestCase):
 
-    def test_recipients_are_expanded_from_group_members(self):
-        storage = mock.MagicMock()
-        collection_record = {
+    def setUp(self):
+        self.storage = mock.MagicMock()
+        self.collection_record = {
             'kinto-emailer': {
                 'hooks': [{
                     'template': 'Poll changed.',
@@ -221,14 +264,41 @@ class GroupExpansionTest(unittest.TestCase):
                 }]
             }
         }
-        group_record = {
+        self.group_record = {
             'members': ['fxa:225689876', 'portier:devnull@localhost.com']
         }
-        storage.get.side_effect = [collection_record, group_record]
+        self.storage.get.side_effect = [self.collection_record, self.group_record]
+
+    def test_recipients_are_expanded_from_group_members(self):
         payload = {'bucket_id': 'b', 'collection_id': 'c'}
-        message, = get_messages(storage, payload)
+        message, = get_messages(self.storage, payload)
         assert message.recipients == ['devnull@localhost.com']
 
+    def test_no_message_no_email_in_group_members(self):
+        group_record = {
+            'members': ['fxa:225689876', 'basicauth:toto-la-mamposina']
+        }
+        self.storage.get.side_effect = [self.collection_record, group_record]
+        payload = {'bucket_id': 'b', 'collection_id': 'c'}
+        messages = get_messages(self.storage, payload)
+        assert not messages
+
+    def test_email_group_can_contain_placeholders(self):
+        collection_record = {
+            'kinto-emailer': {
+                'hooks': [{
+                    'template': 'Poll changed.',
+                    'recipients': ['/buckets/b/groups/{collection_id}'],
+                }]
+            }
+        }
+        self.storage.get.side_effect = [collection_record, self.group_record]
+        payload = {'bucket_id': 'b', 'collection_id': 'c'}
+        get_messages(self.storage, payload)
+        self.storage.get.assert_called_with(parent_id='/buckets/b',
+                                            collection_id='group',
+                                            object_id='c')
+
 
 class SendNotificationTest(unittest.TestCase):
     def test_send_notification_does_not_call_the_mailer_if_no_message(self):
@@ -243,10 +313,11 @@ class SendNotificationTest(unittest.TestCase):
 
         with mock.patch('kinto_emailer.get_mailer') as get_mailer:
             send_notification(event)
-            assert not get_mailer().send.called
+            assert not get_mailer().send_immediately.called
 
     def test_send_notification_calls_the_mailer_if_match_event(self):
         event = mock.MagicMock()
+        event.impacted_records = [{'new': {'id': 'a'}}]
         event.payload = {
             'resource_name': 'record',
             'action': 'update',
@@ -254,15 +325,33 @@ class SendNotificationTest(unittest.TestCase):
             'collection_id': 'foobar'
         }
         event.request.registry.storage.get.return_value = COLLECTION_RECORD
+        event.request.registry.settings = {}
 
         with mock.patch('kinto_emailer.get_mailer') as get_mailer:
             send_notification(event)
-            assert get_mailer().send.called
+            assert get_mailer().send_immediately.called
+
+    def test_send_notification_calls_the_mailer_queue_if_configured(self):
+        event = mock.MagicMock()
+        event.impacted_records = [{'new': {'id': 'a'}}]
+        event.payload = {
+            'resource_name': 'record',
+            'action': 'update',
+            'bucket_id': 'default',
+            'collection_id': 'foobar'
+        }
+        event.request.registry.storage.get.return_value = COLLECTION_RECORD
+        event.request.registry.settings = {'mail.queue_path': '/var/mail'}
+
+        with mock.patch('kinto_emailer.get_mailer') as get_mailer:
+            send_notification(event)
+            assert get_mailer().send_to_queue.called
 
 
 class SignerEventsTest(EmailerTest):
-    def get_app_settings(self, extras=None):
-        settings = super(SignerEventsTest, self).get_app_settings(extras)
+    @classmethod
+    def get_app_settings(cls, extras=None):
+        settings = super().get_app_settings(extras)
         settings['kinto.includes'] += ' kinto_signer'
         settings['kinto.signer.resources'] = (
             '/buckets/staging/collections/addons;'
@@ -276,6 +365,7 @@ class SignerEventsTest(EmailerTest):
         return settings
 
     def setUp(self):
+        self._patch_autograph()
         collection = {
             'kinto-emailer': {
                 'hooks': [{
@@ -293,13 +383,14 @@ class SignerEventsTest(EmailerTest):
         self.app.post_json('/buckets/staging/collections/addons/records',
                            {'data': {'age': 42}},
                            headers=self.headers)
-        self._patch_autograph()
 
     def _patch_autograph(self):
         # Patch calls to Autograph.
         patch = mock.patch('kinto_signer.signer.autograph.requests')
+        self.addCleanup(patch.stop)
         mocked = patch.start()
         mocked.post.return_value.json.return_value = [{
+            "ref": "",
             "signature": "",
             "hash_algorithm": "",
             "signature_encoding": "",
@@ -312,7 +403,41 @@ class SignerEventsTest(EmailerTest):
             self.app.patch_json('/buckets/staging/collections/addons',
                                 {'data': {'status': 'to-review'}},
                                 headers=self.headers)
-            assert get_mailer().send.called
+            assert get_mailer().send_immediately.called
+
+
+class BatchRequestTest(EmailerTest):
+    def setUp(self):
+        bucket = {
+            'kinto-emailer': {
+                'hooks': [{
+                    'action': 'create',
+                    'resource_name': 'collection',
+                    'subject': 'Created {bucket_id}/{collection_id}.',
+                    'template': '',
+                    'recipients': ['me@you.com'],
+                }]
+            }
+        }
+        self.headers = dict(self.headers, **get_user_headers('nous'))
+        self.app.put_json('/buckets/b', {'data': bucket}, headers=self.headers)
+
+        patch = mock.patch('kinto_emailer.get_mailer')
+        self.get_mailer = patch.start()
+        self.addCleanup(patch.stop)
+
+    def test_emails_are_generated_for_each_batch_subrequest(self):
+        requests = {
+            'defaults': {'method': 'POST', 'path': '/buckets/b/collections'},
+            'requests': [
+                {'body': {'data': {'id': '1'}}},
+                {'body': {'data': {'id': '2'}}}
+            ]
+        }
+        self.app.post_json('/batch', requests, headers=self.headers)
+        call1, call2 = self.get_mailer().send_immediately.call_args_list
+        assert call1[0][0].subject == "Created b/1."
+        assert call2[0][0].subject == "Created b/2."
 
 
 class HookValidationTest(FormattedErrorMixin, EmailerTest):
@@ -365,13 +490,26 @@ class HookValidationTest(FormattedErrorMixin, EmailerTest):
                               status=400)
         assert 'Missing "hooks"' in r.json['message']
 
-    def test_validation_errors_are_well_formatted(self):
+    def test_bucket_validation_errors_are_well_formatted(self):
+        self.valid_collection['kinto-emailer'].pop('hooks')
+        r = self.app.post_json('/buckets',
+                               {'data': self.valid_collection},
+                               headers=self.headers,
+                               status=400)
+        self.assertFormattedError(r, 400,
+                                  errno=errors.ERRORS.INVALID_PARAMETERS,
+                                  error='Invalid parameters',
+                                  message='Missing "hooks"', info=None)
+
+    def test_collection_validation_errors_are_well_formatted(self):
         self.valid_collection['kinto-emailer'].pop('hooks')
         r = self.app.put_json('/buckets/b/collections/c',
                               {'data': self.valid_collection},
                               headers=self.headers,
                               status=400)
-        self.assertFormattedError(r, 400, errno=107, error='Invalid parameters',
+        self.assertFormattedError(r, 400,
+                                  errno=errors.ERRORS.INVALID_PARAMETERS,
+                                  error='Invalid parameters',
                                   message='Missing "hooks"', info=None)
 
     def test_fails_if_missing_template(self):
@@ -420,12 +558,3 @@ class HookValidationTest(FormattedErrorMixin, EmailerTest):
                               headers=self.headers,
                               status=400)
         assert 'Invalid recipients /buckets/b/group/g' in r.json['message']
-
-    def test_fails_if_template_contains_bad_placeholders(self):
-        tpl = 'The current temperature is {degres}C.'
-        self.valid_collection['kinto-emailer']['hooks'][0]['template'] = tpl
-        r = self.app.put_json('/buckets/b/collections/c',
-                              {'data': self.valid_collection},
-                              headers=self.headers,
-                              status=400)
-        assert 'degres' in r.json['message']
